import cherrypy
import os
from Framework.PluginManager import PluginManager
from Framework.Context import Context
from Framework.Logger import Logger
import traceback
import sys

class HttpsOnlyFilter (object):
    def __init__ (self, redirect):
        self.redirect = redirect
    def __call__ (self):
        if (cherrypy.request.query_string) and (not "https" in cherrypy.request.query_string):
            raise cherrypy.HTTPRedirect (self.redirect)

class BonsaiServer (object):
    def __init__ (self, context):
        self.context = context
        self.__addCommandLineOptions ()
        self.context.addService (PluginManager ("bonsai.cfg"))
        self.controllers = []
        self.services = []
        self.logger = Logger ("Server")
        self.context.addService (self.logger)
        
        self.__loadServices ()
        self.__loadControllers ()
        self.mountControllers ()
    
    def __addCommandLineOptions (self):
        parser = self.context.OptionParser ()
        parser.add_option ("-s", "--secure", 
                           help="redirects to URL if the connection is not secure", 
                           dest="secureRedirect",
                           metavar="URL",
                           default="")
        
        parser.add_option ("-p", "--port",
                           help="server listening on port PORT",
                           dest="port",
                           metavar="PORT",
                           default="8030")
        
        parser.add_option ("-e", "--environment",
                           help="server environment (production|development)",
                           dest="environment",
                           metavar="ENV",
                           default="production")
        
        parser.add_option ("--default-page",
                           dest="defaultPage",
                           metavar="URL",
                           default="/Studio/login")
        def stripTrailingSlash (option, opt_str, value, parser, *args, **kwargs):
            setattr(parser.values, option.dest, value.rstrip ("/"))
        
        parser.add_option ("--base-url",
                           help="Base URL for the server (for usage behind a proxy).",
                           default="http://localhost:8030",
                           dest="baseUrl",
                           action="callback",
                           callback=stripTrailingSlash,
                           type="str",
                           nargs=1)
    
    def __loadServices (self):
        for plugin in self.context.PluginManager ().plugins ("/Services"):
            # Services are loaded before /Controllers to provide them with
            # general utilities.
            serviceName = plugin.implementation.__name__
            self.logger.debug ("Loading service: %s" % serviceName)
            try:
                service = plugin.createInstance (self.context)
                self.services.append (service)
            except Exception, e:
                message = "An error was found while loading the service: %s.\n" % serviceName
                exceptionStrings = traceback.format_exception (sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.logger.debug (message + "\t".join (exceptionStrings))
                self.logger.warning ("Service %s skipped." % serviceName)

    def __loadControllers (self):
        for plugin in self.context.PluginManager ().plugins ("/Controllers"):
            # Each plugin lives in his own context, with his own logger.
            pluginContext = Context (self.context)
            name = plugin.implementation.__name__
            self.logger.message ("Controller %s loaded." % name)
            pluginContext.addService (Logger (name))
            try:
                controller = plugin.createInstance (pluginContext)
                controller.options = plugin.options
                self.controllers.append (controller)
            except Exception, e:
                message = "An error was found while loading the controller: %s." % name
                exceptionStrings = traceback.format_exception (sys.exc_type, 
                                                               sys.exc_value, 
                                                               sys.exc_traceback)
                self.logger.debug (message + "\t".join (exceptionStrings))
                self.logger.warning ("Controller %s not activated." % name)

    def mountControllers (self):
        for controller in self.controllers:
            cherrypy.tree.mount (controller,
                                 script_name=controller.options["baseUrl"], 
                                 config=controller.config)
        pass

    def start (self):
        (opts, args) = self.context.OptionParser ().parse_args ()
        baseUrl, defaultPage = opts.baseUrl, opts.defaultPage
        class DefaultPage (object):
            def __call__ (*args, **kwds):
                raise cherrypy.HTTPRedirect(baseUrl.strip("/") + defaultPage)
            exposed=True
        cherrypy.tree.mount (DefaultPage ())
        self.context.Logger ().debug ("parsed options: %s" % opts)
        if opts.secureRedirect:
            cherrypy.request.hooks.attach ("before_request_body", HttpsOnlyFilter (opts.secureRedirect))
        for controller in self.controllers:
            controller.readyToRun ()
        cherrypy.config.update ({"server.environment": opts.environment})
        cherrypy.config.update ({"server.socket_port": int (opts.port)})
        cherrypy.server.quickstart ()
        cherrypy.engine.start ()
